from pymongo import MongoClient


class Batcher:

    def __init_connection(self) -> None:
        self.connection = MongoClient()
        self.database = self.connection.get_database(self.database)
        self.collection = self.database.get_collection(f'{self.lang}wiki_raw')

    def __init__(
        self,
        database: str,
        lang: str,
        batch_size: int
    ):
        self.database = database
        self.lang = lang
        self.batch_size = batch_size
        self.__init_connection()
        self.batch_index = 0

    def __iter__(self):
        return self

    def __next__(self):
        data = list(
            self.collection.find({})
                .skip(self.batch_size * self.batch_index)
                .limit(self.batch_size)
        )

        if (data):
            self.batch_index += 1
            return data

        raise StopIteration



from .settings import (
    DEFAULT_DATASETS_DIR,
    DEFAULT_N_PROCESSES,
    DEFAULT_PARALLELIZE, 
    DEFAULT_SYNC_DATA, 
    DEFAULT_LANGUAGE, 
    DEFAULT_DATABASE_PREFIX, 
    DEFAULT_BATCH_SIZE, 
    DEFAULT_FORCE, 
    DEFAULT_SKIP, 
    DEFAULT_ERASE_DATASETS, 
    DEFAULT_DROPOFF_MONTH_THRESHOLD, 
    DEFAULT_METRICS_DIR, 
    DEFAULT_ACTIVE_PER_MONTH_THRESHOLD
)



from pathlib import Path
import multiprocessing

DEFAULT_DATASETS_DIR = Path.home().joinpath('.wikiusers').joinpath('datasets') if __name__ != '__main__' else Path(__file__).parent.parent.parent.joinpath('datasets')
DEFAULT_DATABASE_PREFIX = 'wikiusers'
DEFAULT_LANGUAGE = 'it'
DEFAULT_SYNC_DATA = True
DEFAULT_PARALLELIZE = True
DEFAULT_N_PROCESSES = multiprocessing.cpu_count()
DEFAULT_BATCH_SIZE = int(2e4)
DEFAULT_FORCE = False
DEFAULT_SKIP = False
DEFAULT_ERASE_DATASETS = False


from typing import Union
from pathlib import Path

from wikiusers import logger
from wikiusers import settings
from wikiusers.postprocessor.utils import Batcher, Uploader, elaborate_users_batch, SexElaborator


class PostProcessor:

    def __get_uploader(self, lang: str) -> Uploader:
        uploader = Uploader(self.database, lang, self.force)
        uploader.check_if_collection_already_exists()
        uploader.create_index()
        return uploader

    def __init__(
        self,
        datasets_dir: Union[Path, str] = settings.DEFAULT_DATASETS_DIR,
        database: str = settings.DEFAULT_DATABASE_PREFIX,
        langs: Union[str, list[str]] = settings.DEFAULT_LANGUAGE,
        batch_size: int = settings.DEFAULT_BATCH_SIZE,
        force: bool = settings.DEFAULT_FORCE
    ):
        self.datasets_dir = Path(datasets_dir)
        self.database = database
        self.langs = [langs] if type(langs) == str else langs
        self.batch_size = batch_size
        self.force = force
        self.sex_elaborator: Union[SexElaborator, None] = None

    def process_users(self) -> None:
        for lang in self.langs:
            logger.info('Start postprocessing users', lang=lang, scope='POSTPROCESSOR')
            batcher = Batcher(f'{self.database}_raw', lang, self.batch_size)
            uploader = self.__get_uploader(lang)
            for i, user_batch in enumerate(batcher):
                logger.debug(f'Start processing batch {i}', lang=lang, scope='POSTPROCESSOR')
                processed_users = elaborate_users_batch(user_batch)
                logger.debug(f'Start uploading batch {i}', lang=lang, scope='POSTPROCESSOR')
                uploader.upload_users(processed_users)
                logger.debug(f'Finished batch {i}', lang=lang, scope='POSTPROCESSOR')
            logger.succ('Finished postprocessing users', lang=lang, scope='POSTPROCESSOR')

    def process_sex(self) -> None:
        for lang in self.langs:
            logger.info('Start postprocessing sex', lang=lang, scope='POSTPROCESSOR')

            uploader = self.__get_uploader(lang)

            if not self.sex_elaborator:
                self.sex_elaborator = SexElaborator(self.datasets_dir, lang, self.batch_size)

            for i, user_updates in enumerate(self.sex_elaborator):
                logger.debug(f'Start uploading batch {i}', lang=lang, scope='POSTPROCESSOR')
                uploader.upload_sex(user_updates)
                logger.debug(f'Finished batch {i}', lang=lang, scope='POSTPROCESSOR')

            logger.succ('Finished postprocessing sex', lang=lang, scope='POSTPROCESSOR')

    @staticmethod
    def get_available_langs(dbname: str = settings.DEFAULT_DATABASE_PREFIX) -> list[str]:
        return Uploader.get_available_langs(f'{dbname}_raw')

    @staticmethod
    def get_processed_langs(dbname: str = settings.DEFAULT_DATABASE_PREFIX) -> list[str]:
        return Uploader.get_available_langs(dbname)



from pathlib import Path
from typing import Optional, Tuple, Union
from joblib import Parallel, delayed
from pymongo import MongoClient

from wikiusers import logger
from wikiusers import settings
from wikiusers.dataloader import WhdtLoader
from wikiusers.rawprocessor.utils import Analyzer


class RawProcessor:

    def __init_loader(self, lang: str) -> None:
        loader = WhdtLoader(self.datasets_dir, lang)
        if self.sync_data:
            loader.sync_wikies()
        return loader

    def __check_if_collection_already_exists(self, lang: str) -> bool:
        connection = MongoClient()
        database = connection.get_database(self.database)
        collection = database.get_collection(f'{lang}wiki_raw')

        db_collections = database.list_collection_names()
        if collection.name in db_collections:
            if self.force:
                logger.warn('Collection already exists: dropping',
                            lang=lang, scope='RAWPROCESSOR')
                collection.drop()
            elif self.skip:
                logger.warn('Collection already exists, skipping',
                           lang=lang, scope='RAWPROCESSOR')

                return True
            else:
                logger.warn('Collection already exists, it does not matter',
                           lang=lang, scope='RAWPROCESSOR')
        return False

    def __get_tsv_month_and_year(self, tsv_file_name: str) -> Tuple[Optional[int], Optional[int]]:
        time_str = tsv_file_name.split('.')[-3]
        parts = time_str.split('-')

        if len(parts) > 1:
            [year, month] = parts
        else:
            year = parts[0]

        try:
            month = int(month)
        except:
            month = None

        try:
            year = int(year)
        except:
            year = None

        return month, year

    def __init__(
        self,
        sync_data: bool = settings.DEFAULT_SYNC_DATA,
        datasets_dir: Union[Path, str] = settings.DEFAULT_DATASETS_DIR,
        langs: Union[str, list[str]] = settings.DEFAULT_LANGUAGE,
        parallelize: bool = settings.DEFAULT_PARALLELIZE,
        n_processes: int = settings.DEFAULT_N_PROCESSES,
        database: str = settings.DEFAULT_DATABASE_PREFIX,
        force: bool = settings.DEFAULT_FORCE,
        skip: bool = settings.DEFAULT_SKIP,
        erase_datasets: bool = settings.DEFAULT_ERASE_DATASETS
    ):
        self.sync_data = sync_data
        self.datasets_dir = Path(datasets_dir)
        self.langs = [langs] if type(langs) == str else langs
        self.parallelize = parallelize
        self.n_processes = n_processes
        self.database = f'{database}_raw'
        self.force = force
        self.skip = skip
        self.erase_datasets = erase_datasets

    def _process_file(self, lang: str, path: Path) -> None:
        logger.info(f'Starting processing {path}', lang=lang, scope='RAWPROCESSOR')
        month, year = self.__get_tsv_month_and_year(path.name)
        analyzer = Analyzer(path, month, year, lang, self.database)
        analyzer.analyze()
        logger.succ(f'Finished processing {path}', lang=lang, scope='RAWPROCESSOR')
        if self.erase_datasets:
            path.unlink()
            logger.info(f'Deleted {path}', lang=lang, scope='RAWPROCESSOR')

    def process(self) -> None:
        for lang in self.langs:
            if self.__check_if_collection_already_exists(lang):
                continue

            loader = self.__init_loader(lang)
            datasets_paths = loader.get_tsv_files()

            if self.parallelize:
                Parallel(n_jobs=self.n_processes)(
                    delayed(self._process_file)(lang, path)
                    for path in datasets_paths
                )
            else:
                for path in datasets_paths:
                    self._process_file(lang, path)



from pymongo import ASCENDING, MongoClient
from pymongo.operations import UpdateOne

from wikiusers import logger


class Uploader:

    def __init_connection(self) -> None:
        self.connection = MongoClient()
        self.database = self.connection.get_database(self.database)
        self.collection = self.database.get_collection(f'{self.lang}wiki')

    def __init__(
        self,
        database: str,
        lang: str,
        force: bool
    ):
        self.database = database
        self.lang = lang
        self.force = force
        self.__init_connection()

    def check_if_collection_already_exists(self) -> None:
        db_collections = self.database.list_collection_names()
        if f'{self.lang}wiki' in db_collections:
            if self.force:
                logger.warn('Collection already exists: dropping',
                            lang=self.lang, scope='UPLOADER')
                self.collection.drop()
            else:
                logger.err('Collection already exists',
                           lang=self.lang, scope='UPLOADER')
                raise Exception(f'Collection already exists')

    def create_index(self) -> None:
        self.collection.create_index([('id', ASCENDING)], name='id_index', unique=True)
        logger.debug('Created id index', lang=self.lang, scope='Uploader')

    def upload_users(self, user_batch: list[dict]) -> None:
        self.collection.insert_many(user_batch)
        user_batch.clear()

    def upload_sex(self, user_updates: list[UpdateOne]) -> None:
        self.collection.bulk_write(user_updates)
        user_updates.clear()

    @staticmethod
    def get_available_langs(dbname: str) -> list[str]:
        connection = MongoClient()
        database = connection.get_database(dbname)
        db_collections: list[str] = database.list_collection_names()
        return [collection_name.split('wiki')[0] for collection_name in db_collections]

    



import math
from collections import Counter
from pymongo import MongoClient, UpdateOne, ASCENDING

from wikiusers import logger


class Uploader:
    def __create_index(self) -> None:
        self.collection.create_index([('id', ASCENDING)], name='id_index', unique=True)
        logger.debug('Created id index', lang=self.lang, scope='Uploader')

    def __calc_entropy(self, counter: Counter) -> float:
        values = counter.values()
        sum_of_numbers = sum(value for value in values)
        probs = [value / sum_of_numbers for value in values]
        return -sum([prob * math.log(prob) for prob in probs])

    def __parse_pages_counter(self, pages_counter: Counter) -> dict:
        n = len(pages_counter)
        most_common_5 = dict(pages_counter.most_common(5))
        entropy = self.__calc_entropy(pages_counter)
        return {'n': n, 'most_common_5': most_common_5, 'entropy': entropy}

    def __get_user_inserts(self, user_document: dict) -> list[dict]:
        inserts = []

        for uid in list(user_document.keys()):
            inserts.append(user_document[uid])
            user_document.pop(uid)

        return inserts

    def __get_user_updates(self, user_document_update: dict) -> list[dict]:
        updates = []

        for uid in list(user_document_update.keys()):
            if uid in user_document_update:
                updates.append(UpdateOne({'id': uid}, {'$set': user_document_update[uid]}))
                user_document_update.pop(uid)

        return updates

    def __get_events_updates(self, month: str, year: str, user_month_events: dict, user_helper_info: dict) -> list[dict]:
        updates = []

        for uid in list(user_month_events.keys()):
            month_obj = user_month_events[uid]
            if uid in user_helper_info:
                if 'pages' in user_helper_info[uid]:
                    month_obj['pages'] = self.__parse_pages_counter(user_helper_info[uid]['pages'])
                    user_helper_info[uid].pop('pages')
            updates.append(UpdateOne({'id': uid}, {'$set': {f'events.per_month.{year}.{month}': month_obj}}))
            user_month_events.pop(uid)

        return updates

    def __get_user_alters(self, user_alter_groups: dict, user_alter_blocks: dict) -> list[dict]:
        updates = []

        for uid in list(user_alter_groups.keys()):
            if uid in user_alter_groups:
                updates.append(UpdateOne({'id': uid}, {'$push': {'alter.groups': {'$each': user_alter_groups[uid]}}}))
                user_alter_groups.pop(uid)

        for uid in list(user_alter_blocks.keys()):
            if uid in user_alter_blocks:
                updates.append(UpdateOne({'id': uid}, {'$push': {'alter.blocks': {'$each': user_alter_blocks[uid]}}}))
                user_alter_blocks.pop(uid)

        return updates

    def __get_user_history_usernames(self, user_history_usernames: dict) -> list[dict]:
        updates = []

        for uid in list(user_history_usernames.keys()):
            if uid in user_history_usernames:
                updates.append(
                    UpdateOne({'id': uid}, {'$push': {'history_usernames': {'$each': user_history_usernames[uid]}}}))
                user_history_usernames.pop(uid)

        return updates

    def __init__(self, database: str, lang: str):
        self.lang = lang
        self.connection = MongoClient()
        self.database = self.connection.get_database(database)
        self.collection = self.database.get_collection(f'{lang}wiki_raw')
        self.__create_index()

    def upload_to_db(
        self,
        year: str,
        month: str,
        user_document: dict,
        user_document_update: dict,
        user_month_events: dict,
        user_helper_info: dict,
        user_alter_groups: dict,
        user_alter_blocks: dict,
        user_history_usernames: dict
    ) -> None:
        logger.info(f'Uploading data', year=year, month=month, lang=self.lang, scope='UPLOADER')

        logger.debug(f'Retrieving user inserts', year=year, month=month, lang=self.lang, scope='UPLOADER')
        user_inserts = self.__get_user_inserts(user_document)
        logger.debug(f'Pushing user inserts', year=year, month=month, lang=self.lang, scope='UPLOADER')
        if (user_inserts):
            try:
                self.collection.insert_many(user_inserts, ordered=False)
            except:
                pass
        user_inserts.clear()

        logger.debug(f'Retrieving user updates', year=year, month=month, lang=self.lang, scope='UPLOADER')
        user_updates = self.__get_user_updates(user_document_update)
        logger.debug(f'Pushing user updates', year=year, month=month, lang=self.lang, scope='UPLOADER')
        if (user_updates):
            self.collection.bulk_write(user_updates)
        user_updates.clear()

        logger.debug(f'Retrieving events updates', year=year, month=month, lang=self.lang, scope='UPLOADER')
        events_updates = self.__get_events_updates(month, year, user_month_events, user_helper_info)
        logger.debug(f'Pushing events updates', year=year, month=month, lang=self.lang, scope='UPLOADER')
        if (events_updates):
            self.collection.bulk_write(events_updates)
        events_updates.clear()

        logger.debug(f'Retrieving user alters', year=year, month=month, lang=self.lang, scope='UPLOADER')
        user_alerts = self.__get_user_alters(user_alter_groups, user_alter_blocks)
        logger.debug(f'Pushing user alerts', year=year, month=month, lang=self.lang, scope='UPLOADER')
        if (user_alerts):
            self.collection.bulk_write(user_alerts)
        user_alerts.clear()

        logger.debug(f'Retrieving user history usernames', year=year, month=month, lang=self.lang, scope='UPLOADER')
        user_usernames_updates = self.__get_user_history_usernames(user_history_usernames)
        logger.debug(f'Pushing user history usernames', year=year, month=month, lang=self.lang, scope='UPLOADER')
        if (user_usernames_updates):
            self.collection.bulk_write(user_usernames_updates)
        user_usernames_updates.clear()

        logger.succ(f'Uploading data', year=year, month=month, lang=self.lang, scope='UPLOADER')



def _avg(total: float, count: float) -> float:
    try:
        return total / count
    except ZeroDivisionError:
        return 0


def _fill_basic(input: dict, output: dict) -> None:
    output['id'] = input['id']
    output['username'] = input['username']
    output['creation_timestamp'] = input['creation_timestamp']
    output['registration_timestamp'] = input['registration_timestamp']
    output['is_bot'] = input['is_bot']


def _create_history(alter_groups: list[dict], key: str) -> list[dict]:
    history = []
    for i, alter in enumerate(sorted(alter_groups, key=lambda el: el['timestamp'])):
        timestamp = alter['timestamp']
        history.append({'from': timestamp, 'to': None, key: alter[key]})
        if i > 0:
            history[i - 1]['to'] = timestamp
    return history


def _fill_groups_or_blocks(input: dict, output: dict, key: str) -> None:
    output[key] = {
        'current': input[key],
        'history': _create_history(input['alter'][key], key),
        'ever_had': list({ el for alter in input['alter'][key] for el in alter[key] })
    }


def _fill_usernames_history(input: dict, output: dict) -> None:
    output['usernames_history'] = _create_history(input['history_usernames'], 'username')


def _fill_activity(input: dict, output: dict) -> None:
    activity_per_month = {}
    activity_per_year = {}
    activity_total = {}

    # TOTAL HELPER VARS DECLARATION

    at_max_inactive_interval_in_days_sum = 0
    at_max_inactive_interval_in_days_count = 0
    at_max_inactive_interval_in_days_max = -1
    at_n_activity_days = 0
    at_sesc_since_same_day_event_sum = 0
    at_sesc_since_same_day_event_count = 0
    at_avg_secs_since_last_edit_on_same_page_sum = 0
    at_avg_secs_since_last_edit_on_same_page_count = 0
    at_avg_page_n_sum = 0
    at_avg_page_n_count = 0
    at_avg_page_entropy_sum = 0
    at_avg_page_entropy_count = 0

    events_per_month: dict[str, dict] = input['events']['per_month']
    for year, year_events in sorted(events_per_month.items()):
        activity_per_month[year] = {}

        # PER YEAR HELPER VARS DECLARATION

        activity_per_year[year] = {}
        apy = activity_per_year[year]
        apy_max_inactive_interval_in_days_sum = 0
        apy_max_inactive_interval_in_days_count = 0
        apy_max_inactive_interval_in_days_max = -1
        apy_n_activity_days = 0
        apy_sesc_since_same_day_event_sum = 0
        apy_sesc_since_same_day_event_count = 0
        apy_avg_secs_since_last_edit_on_same_page_sum = 0
        apy_avg_secs_since_last_edit_on_same_page_count = 0
        apy_avg_page_n_sum = 0
        apy_avg_page_n_count = 0
        apy_avg_page_entropy_sum = 0
        apy_avg_page_entropy_count = 0

        for month, month_details in sorted(year_events.items()):
            activity_per_month[year][month] = {}
            apm = activity_per_month[year][month]

            # PER MONTH PROPS

            apm['max_inactive_interval_in_days'] = month_details['max_inact_interval']
            apm['first_event_timestamp'] = month_details['first_event']
            apm['last_event_timestamp'] = month_details['last_event']
            apm['n_of_activity_days'] = month_details['activity_days']
            apm['avg_secs_since_same_day_event'] = _avg(
                month_details['secs_since_same_day_event'],
                month_details['secs_since_same_day_event_count']
            )
            try:
                apm['avg_secs_since_last_edit_on_same_page'] = _avg(
                    month_details['pages_seconds'],
                    month_details['pages_seconds_count']
                )
            except:
                pass

            try:
                apm['pages_activity'] = month_details['pages']
            except:
                pass

            apm['events'] = {'per_namespace': {}, 'total': {}}
            apme = apm['events']
            apme_per_namespace = apme['per_namespace']
            apme_total = apme['total']
            for ns, ns_events in sorted(month_details['namespaces'].items()):
                apme_per_namespace[ns] = {}
                apme_curr_ns = apme_per_namespace[ns]

                total_ns = 0
                for event_type, event_count in sorted(ns_events.items()):
                    apme_curr_ns[event_type] = event_count
                    if event_type != 'minor_edit':
                        total_ns += event_count
                    try:
                        apme_total[event_type] += event_count
                    except:
                        apme_total[event_type] = event_count

                apme_curr_ns['total'] = total_ns
                try:
                    apme_total['total'] += total_ns
                except:
                    apme_total['total'] = total_ns

            # PER YEAR PROPS

            # max inactive interval

            apy_max_inactive_interval_in_days_sum += month_details['max_inact_interval']
            apy_max_inactive_interval_in_days_count += 1
            apy_max_inactive_interval_in_days_max = max(
                apy_max_inactive_interval_in_days_max, month_details['max_inact_interval'])

            # first/last timestamp

            if 'first_event_timestamp' not in apy:
                apy['first_event_timestamp'] = month_details['first_event']

            apy['last_event_timestamp'] = month_details['last_event']

            # n activity days

            apy_n_activity_days += month_details['activity_days']

            # secs since same day event

            apy_sesc_since_same_day_event_sum += month_details['secs_since_same_day_event']
            apy_sesc_since_same_day_event_count += month_details['secs_since_same_day_event_count']

            # secs since last edit on same page

            try:
                apy_avg_secs_since_last_edit_on_same_page_sum += month_details['pages_seconds']
                apy_avg_secs_since_last_edit_on_same_page_count += month_details['pages_seconds_count']
            except:
                pass

            # page n and entropy

            try:
                apy_avg_page_n_sum += month_details['pages']['n']
                apy_avg_page_n_count += 1

                apy_avg_page_entropy_sum += month_details['pages']['entropy']
                apy_avg_page_entropy_count += 1
            except:
                pass

            # TOTAL PROPS

            # first/last timestamp

            if 'first_event_timestamp' not in activity_total:
                activity_total['first_event_timestamp'] = month_details['first_event']

            activity_total['last_event_timestamp'] = month_details['last_event']

        # PER YEAR PROPS

        # max inactive interval

        apy['max_inactive_interval_in_days'] = apy_max_inactive_interval_in_days_max
        apy['avg_max_inactive_interval_in_days'] = _avg(
            apy_max_inactive_interval_in_days_sum, apy_max_inactive_interval_in_days_count)

        # n activity days

        apy['n_of_activity_days'] = apy_n_activity_days

        # secs since same day event

        apy['avg_secs_since_same_day_event'] = _avg(
            apy_sesc_since_same_day_event_sum, apy_sesc_since_same_day_event_count)

        # avg secs since last edit on same page

        apy['avg_secs_since_last_edit_on_same_page'] = _avg(
            apy_avg_secs_since_last_edit_on_same_page_sum, apy_avg_secs_since_last_edit_on_same_page_count)

        # page n and entropy

        apy['pages_activity'] = {
            'avg_n': _avg(
                apy_avg_page_n_sum, apy_avg_page_n_count),
            'avg_entropy': _avg(
                apy_avg_page_entropy_sum, apy_avg_page_entropy_count)
        }

        # events

        apy['events'] = {'per_namespace': {}, 'total': {}}
        apye = apy['events']
        apye_per_namespace = apye['per_namespace']
        apye_total = apye['total']
        for month, month_details in activity_per_month[year].items():
            apme_total = month_details['events']['total']
            apme_per_namespace = month_details['events']['per_namespace']

            for event_type, event_count in apme_total.items():
                try:
                    apye_total[event_type] += event_count
                except:
                    apye_total[event_type] = event_count

            for ns, ns_events in apme_per_namespace.items():
                if ns not in apye_per_namespace:
                    apye_per_namespace[ns] = {}
                apye_curr_ns = apye_per_namespace[ns]

                for event_type, event_count in ns_events.items():
                    try:
                        apye_curr_ns[event_type] += event_count
                    except:
                        apye_curr_ns[event_type] = event_count

        # TOTAL PROPS

        # max inactive interval

        at_max_inactive_interval_in_days_sum += apy_max_inactive_interval_in_days_sum
        at_max_inactive_interval_in_days_count += apy_max_inactive_interval_in_days_count
        at_max_inactive_interval_in_days_max = max(
            at_max_inactive_interval_in_days_max, apy_max_inactive_interval_in_days_max)

        # n activity days

        at_n_activity_days += apy_n_activity_days

        # secs since same day event

        at_sesc_since_same_day_event_sum += apy_sesc_since_same_day_event_sum
        at_sesc_since_same_day_event_count += apy_sesc_since_same_day_event_count

        # avg secs since last edit on same page

        at_avg_secs_since_last_edit_on_same_page_sum += apy_avg_secs_since_last_edit_on_same_page_sum
        at_avg_secs_since_last_edit_on_same_page_count += apy_avg_secs_since_last_edit_on_same_page_count

        # page n and entropy

        at_avg_page_n_sum += apy_avg_page_n_sum
        at_avg_page_n_count += apy_avg_page_n_count

        at_avg_page_entropy_sum += apy_avg_page_entropy_sum
        at_avg_page_entropy_count += apy_avg_page_entropy_count

    # TOTAL PROPS

    # max inactive interval

    activity_total['max_inactive_interval_in_days'] = at_max_inactive_interval_in_days_max
    activity_total['avg_max_inactive_interval_in_days'] = _avg(
        at_max_inactive_interval_in_days_sum, at_max_inactive_interval_in_days_count)

    # n activity days

    activity_total['n_activity_days'] = at_n_activity_days

    # secs since same day event

    activity_total['avg_secs_since_same_day_event'] = _avg(
        at_sesc_since_same_day_event_sum, at_sesc_since_same_day_event_count)

    # avg secs since last edit on same page

    activity_total['avg_secs_since_last_edit_on_same_page'] = _avg(
        at_avg_secs_since_last_edit_on_same_page_sum, at_avg_secs_since_last_edit_on_same_page_count)

    # page n and entropy

    activity_total['pages_activity'] = {
        'avg_n': _avg(
            at_avg_page_n_sum, at_avg_page_n_count),
        'avg_entropy': _avg(
            at_avg_page_entropy_sum, at_avg_page_entropy_count)
    }

    # events

    activity_total['events'] = {'per_namespace': {}, 'total': {}}
    ate = activity_total['events']
    ate_per_namespace = ate['per_namespace']
    ate_total = ate['total']
    for year, year_details in activity_per_year.items():
        apme_total = year_details['events']['total']
        apme_per_namespace = year_details['events']['per_namespace']

        for event_type, event_count in apme_total.items():
            try:
                ate_total[event_type] += event_count
            except:
                ate_total[event_type] = event_count

        for ns, ns_events in apme_per_namespace.items():
            if ns not in ate_per_namespace:
                ate_per_namespace[ns] = {}
            ate_curr_ns = ate_per_namespace[ns]

            for event_type, event_count in ns_events.items():
                try:
                    ate_curr_ns[event_type] += event_count
                except:
                    ate_curr_ns[event_type] = event_count

    # ADD CHAGES TO OUTPUT

    output['activity'] = {
        'total': activity_total,
        'per_year': activity_per_year,
        'per_month': activity_per_month
    }


def _elaborate_user(u: dict) -> dict:
    if u['username']:
        user = {}
        _fill_basic(u, user)
        _fill_groups_or_blocks(u, user, 'groups')
        _fill_groups_or_blocks(u, user, 'blocks')
        _fill_usernames_history(u, user)
        _fill_activity(u, user)
        return user
    else:
        return None


def elaborate_users_batch(users_batch: list[dict]) -> list[dict]:
    elaborated_users = [
        _elaborate_user(user)
        for user in users_batch
    ]
    return [
        user
        for user in elaborated_users
        if user is not None
    ]



import bz2
from pymongo import UpdateOne
from pathlib import Path

from wikiusers import logger

class SexElaborator:

    def __elaborate_sex(self, line: str) -> UpdateOne:
        parts = line.split('\t')
        id = parts[0]
        str_sex = parts[2]
        try:
            sex = True if str_sex == 'male' else (False if str_sex == 'female' else None)
        except:
            logger.warn('Error in parsing sex', parts, lang=self.lang)
        return UpdateOne({'id': int(id)}, {'$set': {'sex': sex}})

    def __init__(self, datasets_dir: Path, lang: str, batch_size: int):
        self.datasets_dir = Path(datasets_dir)
        self.lang = lang
        self.batch_size = batch_size

        path = self.datasets_dir.joinpath('users_sex').joinpath(f'{self.lang}wiki.tsv.bz2')
        self.file = bz2.open(path, 'rt')

    def __iter__(self):
        return self

    def __next__(self):
        batch = []

        for _ in range(self.batch_size):
            if not self.file:
                break

            line = self.file.readline()
            if not line:
                break
            
            batch.append(self.__elaborate_sex(line))

        if (batch):
            return batch

        raise StopIteration



def new_user_insert_obj(uid: int, username: str, creation_timestamp, registration_timestamp, is_bot: bool, groups: list, blocks: list) -> dict:
        return {'id': uid, 'username': username, 'creation_timestamp': creation_timestamp, 'registration_timestamp': registration_timestamp, 'is_bot': is_bot, 'groups': groups, 'blocks': blocks, 'events': {'per_month': {}}}

def new_user_update_obj(uid: int, username: str, creation_timestamp, registration_timestamp, is_bot: bool, groups: list, blocks: list) -> dict:
    return {'id': uid, 'username': username, 'creation_timestamp': creation_timestamp, 'registration_timestamp': registration_timestamp, 'is_bot': is_bot, 'groups': groups, 'blocks': blocks}

def new_month_obj() -> list:
    return {'max_inact_interval': 0}

def two_digits(n: int) -> str:
    return str(n) if n > 9 else '0' + str(n)



from termcolor import colored
from datetime import datetime
from typing import Optional


def _get_asset_str(month: Optional[str], year: Optional[str]) -> str:
    separator = '/' if month and year else ''
    year = '' if year is None else year
    month = '' if month is None else month
    return f'{month}{separator}{year}'


def _log_scopes(lang: Optional[str] = None, year: Optional[str] = None, month: Optional[str] = None, scope: Optional[str] = None) -> str:
    lang_str = colored(f'{{{lang}}}', 'yellow') if lang else ''
    asset_str = colored(f'{{{_get_asset_str(month, year)}}}', 'yellow') if month else ''
    scope_str = colored(f'{{{scope}}}', 'cyan') if scope else ''
    return ' '.join([txt for txt in [lang_str, asset_str, scope_str] if txt])


def _log(*args: list[str], tag: str, colour: str, lang: Optional[str] = None, month: Optional[str] = None, year: Optional[str] = None, scope: Optional[str] = None) -> None:
    timestamp = datetime.now().isoformat()
    timestamp_str = colored(f'[{timestamp}]', 'cyan')
    tag = colored(f'[{tag}]', colour, attrs=['bold'])
    text = _log_scopes(lang=lang, month=month, year=year, scope=scope)
    print(f'{timestamp_str} {tag} {text}', *args, flush=True)


def info(*args: list[str], lang: Optional[str] = None, month: Optional[str] = None, year: Optional[str] = None, scope: Optional[str] = None) -> None:
    _log(*args, lang=lang, month=month, year=year, scope=scope, tag='INFO', colour='blue')


def succ(*args: list[str], lang: Optional[str] = None, month: Optional[str] = None, year: Optional[str] = None, scope: Optional[str] = None) -> None:
    _log(*args, lang=lang, month=month, year=year, scope=scope, tag='SUCC', colour='green')


def debug(*args: list[str], lang: Optional[str] = None, month: Optional[str] = None, year: Optional[str] = None, scope: Optional[str] = None) -> None:
    _log(*args, lang=lang, month=month, year=year, scope=scope, tag='DEBUG', colour='grey')


def warn(*args: list[str], lang: Optional[str] = None, month: Optional[str] = None, year: Optional[str] = None, scope: Optional[str] = None) -> None:
    _log(*args, lang=lang, month=month, year=year, scope=scope, tag='WARN', colour='yellow')


def err(*args: list[str], lang: Optional[str] = None, month: Optional[str] = None, year: Optional[str] = None, scope: Optional[str] = None) -> None:
    _log(*args, lang=lang, month=month, year=year, scope=scope, tag='ERROR', colour='red')



class WhdtKeys:
    wiki_db = 0
    event_entity = 1
    event_type = 2
    event_timestamp = 3
    event_comment = 4
    event_user_id = 5
    event_user_text_historical = 6
    event_user_text = 7
    event_user_blocks_historical = 8
    event_user_blocks = 9
    event_user_groups_historical = 10
    event_user_groups = 11
    event_user_is_bot_by_historical = 12
    event_user_is_bot_by = 13
    event_user_is_created_by_self = 14
    event_user_is_created_by_system = 15
    event_user_is_created_by_peer = 16
    event_user_is_anonymous = 17
    event_user_registration_timestamp = 18
    event_user_creation_timestamp = 19
    event_user_first_edit_timestamp = 20
    event_user_revision_count = 21
    event_user_seconds_since_previous_revision = 22
    page_id = 23
    page_title_historical = 24
    page_title = 25
    page_namespace_historical = 26
    page_namespace_is_content_historical = 27
    page_namespace = 28
    page_namespace_is_content = 29
    page_is_redirect = 30
    page_is_deleted = 31
    page_creation_timestamp = 32
    page_first_edit_timestamp = 33
    page_revision_count = 34
    page_seconds_since_previous_revision = 35
    user_id = 36
    user_text_historical = 37
    user_text = 38
    user_blocks_historical = 39
    user_blocks = 40
    user_groups_historical = 41
    user_groups = 42
    user_is_bot_by_historical = 43
    user_is_bot_by = 44
    user_is_created_by_self = 45
    user_is_created_by_system = 46
    user_is_created_by_peer = 47
    user_is_anonymous = 48
    user_registration_timestamp = 49
    user_creation_timestamp = 50
    user_first_edit_timestamp = 51
    revision_id = 52
    revision_parent_id = 53
    revision_minor_edit = 54
    revision_deleted_parts = 55
    revision_deleted_parts_are_suppressed = 56
    revision_text_bytes = 57
    revision_text_bytes_diff = 58
    revision_text_sha1 = 59
    revision_content_model = 60
    revision_content_format = 61
    revision_is_deleted_by_page_deletion = 62
    revision_deleted_by_page_deletion_timestamp = 63
    revision_is_identity_reverted = 64
    revision_first_identity_reverting_revision_id = 65
    revision_seconds_to_identity_revert = 66
    revision_is_identity_revert = 67
    revision_is_from_before_page_creation = 68
    revision_tags = 69


EVENTS_MAP = {
    'create': 'create',
    'create-page': 'create',
    'delete': 'delete',
    'restore': 'restore',
    'move': 'move',
    'merge': 'merge',
    'edit': 'edit'
}



import requests
from typing import Union
from pathlib import Path
from whdtscraper import fetch_wikies, fetch_dumps, fetch_latest_version, WIKI_URL

from wikiusers import logger


class WhdtLoader:

    def __init__(self, datasets_dir: Union[str, Path], lang: str):
        datasets_dir = Path(datasets_dir)
        self.wiki_dir = datasets_dir.joinpath('whdt')
        self.lang = lang

    def __download(self, file_path: str, file_path_obj: Path, tsv_url: str) -> None:
        file_path_obj.parent.mkdir(exist_ok=True, parents=True)
        response = requests.get(tsv_url, allow_redirects=True)
        open(file_path_obj, 'wb').write(response.content)
        logger.debug(f'Downloaded {file_path}', lang=self.lang, scope='scraper')

    def __download_if_needed(self, file_path: str, size_in_bytes: int, tsv_url: str) -> None:
        logger.debug(f'Syncing {file_path}', lang=self.lang, scope='scraper')
        file_path_obj = self.wiki_dir.joinpath(file_path)

        if not file_path_obj.is_file():
            self.__download(file_path, file_path_obj, tsv_url)
        elif abs(file_path_obj.stat().st_size - size_in_bytes) > 50e6:
            logger.warn(f'{file_path} already exists but with unexpected size, redownloading',
                        lang=self.lang, scope='scraper')
            file_path_obj.unlink()
            self.__download(file_path, file_path_obj, tsv_url)
        else:
            logger.debug(f'{file_path} already exists', lang=self.lang, scope='scraper')

    def sync_wikies(self, /, version: str = 'latest'):
        logger.info('Syncing tsv files', lang=self.lang, scope='scraper')

        if (version == 'latest'):
            last_version_obj = fetch_latest_version()
            version = last_version_obj['version']

        wiki = f'{self.lang}wiki'
        dumps = fetch_dumps(version, wiki)

        for dump in dumps:
            url: str = dump['url']
            size_in_bytes: int = int(dump['bytes'])
            self.__download_if_needed(url.replace(f'{WIKI_URL}/', ''), size_in_bytes, url)

        logger.succ(f'Synced tsv files', lang=self.lang, scope='scraper')

    def get_tsv_files(self, /, version: str = 'latest') -> list[Path]:
        if version == 'latest':
            version_path = list(self.wiki_dir.glob('*'))[-1]
        else:
            version_path = self.wiki_dir.joinpath(version)

        wiki = f'{self.lang}wiki'
        wiki_path = version_path.joinpath(wiki)
        files = [file for file in wiki_path.iterdir() if file.is_file()]
        return sorted(files, key=lambda f: f.stem)

    def get_available_langs(self) -> list[str]:
        unfiltered_langs = sorted([wiki['wiki'].replace('wiki', '') for wiki in fetch_wikies('latest', wikitype='wiki')])
        return [lang for lang in unfiltered_langs if len(lang) == 2]

    def get_local_langs(self) -> list[str]:
        versions = sorted([dir.name for dir in self.wiki_dir.iterdir() if dir.is_dir()], reverse=True)
        if not versions:
            return []

        path = self.wiki_dir.joinpath(versions[0])
        return sorted([dir.name.replace('wiki', '') for dir in path.iterdir() if dir.is_dir()])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['wikiusers',
 'wikiusers.dataloader',
 'wikiusers.logger',
 'wikiusers.postprocessor',
 'wikiusers.postprocessor.utils',
 'wikiusers.postprocessor.utils.elaborator',
 'wikiusers.rawprocessor',
 'wikiusers.rawprocessor.utils',
 'wikiusers.settings']

package_data = \
{'': ['*']}

install_requires = \
['autopep8>=1.5.7,<2.0.0',
 'joblib>=1.0.1,<2.0.0',
 'pylint>=2.9.3,<3.0.0',
 'pymongo>=3.11.4,<4.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'termcolor>=1.1.0,<2.0.0',
 'whdtscraper>=1.0.1,<2.0.0']

setup_kwargs = {
    'name': 'wikiusers',
    'version': '1.1.0',
    'description': 'A python3 module that uses the Mediawiki History Dump Tsv to download the datasets, analyze them, uploading them to MongoDB and postprocess them. The result is a database with tons of information on every individual user of Wikipedia',
    'long_description': None,
    'author': 'Eugenio Vinicio Berretta',
    'author_email': 'euberdeveloper@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/euberdeveloper/wikiusers',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



from .keys import WhdtKeys, EVENTS_MAP
from .parser import parse_date, parse_int, parse_bool, parse_str_array
from .uploader import Uploader
from .analyzer import Analyzer
from .misc import new_month_obj, new_user_insert_obj, new_user_update_obj, two_digits


from typing import Union
from datetime import datetime
from dateutil import parser as date_parser


def parse_date(date: str) -> Union[datetime, None]:
    return date_parser.parse(date) if date != '' else None


def parse_int(num: str) -> Union[int, None]:
    return int(num) if num != '' else None

def parse_bool(val: str) -> bool:
    return val != ''

def parse_str_array(arr: str) -> list[str]:
    return arr.split(',') if arr != '' else []



import bz2
from collections import Counter
from datetime import datetime
from pathlib import Path
from typing import Optional

from wikiusers import logger
from wikiusers.rawprocessor.utils import Uploader,  WhdtKeys,  EVENTS_MAP
from .misc import two_digits, new_month_obj, new_user_insert_obj, new_user_update_obj
from .parser import parse_date, parse_int, parse_bool, parse_str_array


class Analyzer:
    def __init_globals(self) -> None:
        self.user_exists = set()
        self.user_document = {}
        self.user_document_update = {}
        self.user_month_events = {}
        self.user_alter_groups = {}
        self.user_alter_blocks = {}
        self.user_history_usernames = {}
        self.user_helper_info = {}

    def __check_if_new_month(self, timestamp: datetime, check: bool) -> None:
        month = two_digits(timestamp.month)
        # assumes not changes such as 04/2001 -> 04/2002
        if (not check or month != self.current_month):
            year = str(timestamp.year)
            logger.info(f'New month {month}/{year}', lang=self.lang, scope='ANALYZER')
            self.uploader.upload_to_db(
                self.current_year,
                self.current_month,
                self.user_document,
                self.user_document_update,
                self.user_month_events,
                self.user_helper_info,
                self.user_alter_groups,
                self.user_alter_blocks,
                self.user_history_usernames
            )
            self.current_month = month
            self.current_year = year

    def __update_month_object(self, uid: str, namespace: str, event_type: str, timestamp: datetime, minor_edit: bool, page_id: str, page_seconds_since_previous_revision: str) -> None:
        if uid not in self.user_month_events:
            self.user_month_events[uid] = new_month_obj()
            user_month = self.user_month_events[uid]
            user_month['first_event'] = timestamp
            user_month['activity_days'] = 1
        else:
            user_month = self.user_month_events[uid]

        if 'namespaces' not in user_month:
            user_month['namespaces'] = {}

        user_month_namespaces = user_month['namespaces']

        if namespace not in user_month_namespaces:
            user_month_namespaces[namespace] = {}

        if uid not in self.user_helper_info:
            self.user_helper_info[uid] = {}
        user_helper_info = self.user_helper_info[uid]

        user_month_namespace = user_month_namespaces[namespace]
        if event_type not in user_month_namespace:
            user_month_namespace[event_type] = 1
        else:
            user_month_namespace[event_type] += 1
        if minor_edit:
            if 'minor_edit' not in user_month_namespace:
                user_month_namespace['minor_edit'] = 1
            else:
                user_month_namespace['minor_edit'] += 1

        if 'last_event' in user_month:
            last_timestamp = user_month['last_event']
            day_diff = timestamp.day - last_timestamp.day
            if day_diff == 0:
                user_month['secs_since_same_day_event'] += (timestamp - last_timestamp).total_seconds()
                user_month['secs_since_same_day_event_count'] += 1
            else:
                user_month['activity_days'] += 1
                if day_diff > user_month['max_inact_interval']:
                    user_month['max_inact_interval'] = day_diff

        else:
            user_month['secs_since_same_day_event'] = 0
            user_month['secs_since_same_day_event_count'] = 0

        if page_id != '':
            if 'pages' not in user_helper_info:
                user_helper_info['pages'] = Counter()
            user_helper_info['pages'].update([page_id])

        if page_seconds_since_previous_revision is not None:
            if 'pages_seconds' not in user_month:
                user_month['pages_seconds'] = page_seconds_since_previous_revision
                user_month['pages_seconds_count'] = 1
            else:
                user_month['pages_seconds'] += page_seconds_since_previous_revision
                user_month['pages_seconds_count'] += 1

        user_month['last_event'] = timestamp

    def __add_provvisory_user_insert(self, uid: str, parts: list[str]) -> None:
        if uid not in self.user_exists:
            username = parts[WhdtKeys.user_text]
            creation_timestamp = parse_date(parts[WhdtKeys.user_creation_timestamp])
            registration_timestamp = parse_date(parts[WhdtKeys.user_registration_timestamp])
            is_bot = parts[WhdtKeys.user_is_bot_by] != ''
            groups = parse_str_array(parts[WhdtKeys.user_groups])
            blocks = parse_str_array(parts[WhdtKeys.user_blocks])
            self.user_document[uid] = new_user_insert_obj(
                uid, username, creation_timestamp, registration_timestamp, is_bot, groups, blocks)
            self.user_exists.add(uid)

    def __add_current_user_groups(self, uid: str, current_groups: list[str], timestamp: datetime) -> None:
        if uid not in self.user_alter_groups:
            self.user_alter_groups[uid] = [{'timestamp': timestamp, 'groups': current_groups}]
        else:
            self.user_alter_groups[uid].append({'timestamp': timestamp, 'groups': current_groups})

    def __add_current_user_blocks(self, uid: str, current_blocks: list[str], timestamp: datetime) -> None:
        if uid not in self.user_alter_blocks:
            self.user_alter_blocks[uid] = [{'timestamp': timestamp, 'blocks': current_blocks}]
        else:
            self.user_alter_blocks[uid].append({'timestamp': timestamp, 'blocks': current_blocks})

    def __add_current_user_username(self, uid: str, current_username: list[str], timestamp: datetime) -> None:
        if uid not in self.user_history_usernames:
            self.user_history_usernames[uid] = [{'timestamp': timestamp, 'username': current_username}]
        else:
            self.user_history_usernames[uid].append({'timestamp': timestamp, 'username': current_username})

    def __analyze_user_create(self, parts: list[str], timestamp: datetime) -> None:
        uid = parse_int(parts[WhdtKeys.user_id])

        if uid is not None:
            username = parts[WhdtKeys.user_text]
            creation_timestamp = parse_date(parts[WhdtKeys.user_creation_timestamp])
            registration_timestamp = parse_date(parts[WhdtKeys.user_registration_timestamp])
            is_bot = parse_bool(parts[WhdtKeys.user_is_bot_by])
            groups = parse_str_array(parts[WhdtKeys.user_groups])
            blocks = parse_str_array(parts[WhdtKeys.user_blocks])
            self.user_document[uid] = new_user_insert_obj(uid, username, creation_timestamp,
                                                          registration_timestamp, is_bot, groups, blocks)
            self.user_document_update[uid] = new_user_update_obj(
                uid, username, creation_timestamp, registration_timestamp, is_bot, groups, blocks)

            current_groups = parse_str_array(parts[WhdtKeys.user_groups_historical])
            self.__add_current_user_groups(uid, current_groups, timestamp)

            current_blocks = parse_str_array(parts[WhdtKeys.user_blocks_historical])
            self.__add_current_user_blocks(uid, current_blocks, timestamp)

            current_username = parts[WhdtKeys.user_text_historical]
            self.__add_current_user_username(uid, current_username, timestamp)

    def __analyze_user_altergroups(self, parts: list[str], timestamp: datetime) -> None:
        uid = parse_int(parts[WhdtKeys.user_id])
        if uid is not None:
            self.__add_provvisory_user_insert(uid, parts)

            current_groups = parse_str_array(parts[WhdtKeys.user_groups_historical])
            self.__add_current_user_groups(uid, current_groups, timestamp)

    def __analyze_user_alterblocks(self, parts: list[str], timestamp: datetime) -> None:
        uid = parse_int(parts[WhdtKeys.user_id])
        if uid is not None:
            self.__add_provvisory_user_insert(uid, parts)

            current_blocks = parse_str_array(parts[WhdtKeys.user_blocks_historical])
            self.__add_current_user_blocks(uid, current_blocks, timestamp)

    def __analyze_user_rename(self, parts: list[str], timestamp: datetime) -> None:
        uid = parse_int(parts[WhdtKeys.user_id])
        if uid is not None:
            self.__add_provvisory_user_insert(uid, parts)

            current_username = parts[WhdtKeys.user_text_historical]
            self.__add_current_user_username(uid, current_username, timestamp)

    def __analyze_page_or_revision(self, event_type: str, timestamp: datetime, parts: list[str]) -> None:
        uid = parse_int(parts[WhdtKeys.event_user_id])

        if uid is not None:
            if uid not in self.user_exists:
                username = parts[WhdtKeys.event_user_text]
                creation_timestamp = parse_date(parts[WhdtKeys.event_user_creation_timestamp])
                registration_timestamp = parse_date(parts[WhdtKeys.event_user_registration_timestamp])
                is_bot = parts[WhdtKeys.event_user_is_bot_by] != ''
                groups = parse_str_array(parts[WhdtKeys.event_user_groups])
                blocks = parse_str_array(parts[WhdtKeys.event_user_blocks])
                self.user_document[uid] = new_user_insert_obj(
                    uid, username, creation_timestamp, registration_timestamp, is_bot, groups, blocks)
                self.user_exists.add(uid)

            minor_edit = parts[WhdtKeys.revision_minor_edit] == 'true'
            page_id = parts[WhdtKeys.page_id]
            page_seconds_since_previous_revision = parse_int(parts[WhdtKeys.page_seconds_since_previous_revision])
            namespace = parse_int(parts[WhdtKeys.page_namespace])
            namespace = f'n{namespace}' if namespace is not None else 'unknown'
            self.__update_month_object(uid, namespace, EVENTS_MAP[event_type], timestamp,
                                       minor_edit, page_id, page_seconds_since_previous_revision)

    def __init__(
        self,
        path: Path,
        start_month: Optional[int],
        start_year: Optional[int],
        lang: str,
        database: str
    ):
        self.path = path
        self.lang = lang

        self.current_month = '01' if start_month is None else two_digits(start_month)
        self.current_year = str(start_year) if start_year is not None else start_year

        self.uploader = Uploader(database, lang)
        self.__init_globals()

    def analyze(self) -> None:
        if self.current_year is None:
            with bz2.open(self.path, 'rt') as input:
                line = input.readline()
                parts = line.split('\t')
                timestamp = parse_date(parts[WhdtKeys.event_timestamp])
                self.current_year = str(timestamp.year)

        with bz2.open(self.path, 'rt') as input:
            timestamp = None
            for line in input:
                parts = line.split('\t')
                timestamp = parse_date(parts[WhdtKeys.event_timestamp])
                self.__check_if_new_month(timestamp, True)

                event_entity = parts[WhdtKeys.event_entity]
                event_type = parts[WhdtKeys.event_type]

                if event_entity == 'revision':
                    self.__analyze_page_or_revision('edit', timestamp, parts)
                elif event_entity == 'page':
                    self.__analyze_page_or_revision(event_type, timestamp, parts)
                elif event_entity == 'user':
                    if event_type == 'create':
                        self.__analyze_user_create(parts, timestamp)
                    elif event_type == 'altergroups':
                        self.__analyze_user_altergroups(parts, timestamp)
                    elif event_type == 'alterblocks':
                        self.__analyze_user_alterblocks(parts, timestamp)
                    elif event_type == 'rename':
                        self.__analyze_user_rename(parts, timestamp)

            self.__check_if_new_month(timestamp, False)
